// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

#include "subsystems/Chassis.h"
#include "Units.h"

#include <frc/kinematics/SwerveModuleState.h>
#include <frc/smartdashboard/SmartDashboard.h>
#include <units/angle.h>

Chassis::Chassis() {
  // Implementation of subsystem constructor goes here.
  zero_pod_encoders();

  configure_pod(m_lf_pod);
  configure_pod(m_rf_pod);
  configure_pod(m_lb_pod);
  configure_pod(m_rb_pod);
}

void Chassis::zero_pod_encoders() {
  m_lf_pod.turn.SetSelectedSensorPosition(0);
  m_rf_pod.turn.SetSelectedSensorPosition(0);
  m_lb_pod.turn.SetSelectedSensorPosition(0);
  m_rb_pod.turn.SetSelectedSensorPosition(0);
}

void Chassis::zero_gyro() {
  // m_gyro.ZeroYaw();
}

void Chassis::drive(frc::ChassisSpeeds speeds) {
  auto module_states = m_kinematics.ToSwerveModuleStates(speeds);
  m_kinematics.DesaturateWheelSpeeds(&module_states, constants::chassis::MAX_SPEED);

  auto lf_optimized = frc::SwerveModuleState::Optimize(
    module_states[0],
    get_pod_angle(m_lf_pod));
  auto rf_optimized = frc::SwerveModuleState::Optimize(
    module_states[1],
    get_pod_angle(m_rf_pod));
  auto lb_optimized = frc::SwerveModuleState::Optimize(
    module_states[2],
    get_pod_angle(m_lb_pod));
  auto rb_optimized = frc::SwerveModuleState::Optimize(
    module_states[3],
    get_pod_angle(m_rb_pod));

  set_pod(m_lf_pod, lf_optimized);
  set_pod(m_rf_pod, rf_optimized);
  set_pod(m_lb_pod, lb_optimized);
  set_pod(m_rb_pod, rb_optimized);
}

void Chassis::drive_field_relative(
    units::meters_per_second_t forward,
    units::meters_per_second_t left,
    units::radians_per_second_t theta) {
  auto speeds = frc::ChassisSpeeds::FromFieldRelativeSpeeds(
    forward,
    left,
    theta,
    // frc::Rotation2d(units::degree_t(-m_gyro.GetAngle()))
    frc::Rotation2d(units::degree_t(0))
  );

  drive(speeds);
}

void Chassis::configure_pod(Pod &pod) {
  pod.turn.ConfigSelectedFeedbackSensor(TalonSRXFeedbackDevice::CTRE_MagEncoder_Relative);
  pod.turn.SetSensorPhase(true);
  pod.turn.Config_kP(0, constants::chassis::TURN_P);
  pod.turn.Config_kI(0, constants::chassis::TURN_I);
  pod.turn.Config_kD(0, constants::chassis::TURN_D);

  pod.drive.ConfigSupplyCurrentLimit({true, 15, 0, 0.01});
}

void Chassis::set_pod(Pod &pod, frc::SwerveModuleState command) {
    // TODO: I can't remember if we waited for the angle target to be met before we
    // start to apply the drive velocity, look into that
    double drive_percent = command.speed / constants::chassis::MAX_SPEED;
    units::ctre_count_t counts =
      command.angle.Degrees() * constants::chassis::ROTATE_RATIO;

    std::string pod_name;
    if(&pod == &m_lf_pod)
      pod_name = "lf_pod";
    if(&pod == &m_rf_pod)
      pod_name = "rf_pod";
    if(&pod == &m_lb_pod)
      pod_name = "lb_pod";
    if(&pod == &m_rb_pod)
      pod_name = "rb_pod";
    frc::SmartDashboard::PutNumber(pod_name + ".angle", command.angle.Degrees().value());
    frc::SmartDashboard::PutNumber(pod_name + ".mag", drive_percent);

    if(abs(drive_percent) > constants::chassis::MIN_DRIVE_PERCENT)
      pod.turn.Set(ControlMode::Position, counts.value());
    pod.drive.Set(ControlMode::PercentOutput, drive_percent);
    pod.tmp = command.angle.Degrees().value();
}

frc::Rotation2d Chassis::get_pod_angle(Pod &pod) {
  units::ctre_count_t counts{pod.turn.GetSelectedSensorPosition()};
  units::radian_t angle{counts / constants::chassis::ROTATE_RATIO};
  return frc::Rotation2d{units::degree_t{pod.tmp}};

  //return frc::Rotation2d{angle};
}
